#include <stdio.h>
#include <ctype.h>
#include <stdlib.h>
#include "pvr.h"

static int fill_int(char *name, char *str, void *ptr, void *ctl);
static int fill_flag(char *name, char *str, void *ptr, void *ctl);
static int fill_string(char *name, char *str, void *ptr, void *ctl);
static int fill_enum(char *name, char *str, void *ptr, void *ctl);

/* fill configuration variables with default values */
cfgstruct config = {
	.path = "",
	.target1 = "/dtv/usb/sda1",
	.target2 = "/mtd_wiselink/MOVIE001",
	.debug = 1,
	.exedsp_debug = 0,
	.default_source = 0x0,
	.aspect_ratio = 1,
	.record_key_lock = 1,
	.protect = 1,
	.unicode = CFG_UNICODE_UTF8,
	.winter_time = 1,
	.summer_time = 2,
	.show_icons = 1,
	.icon_x = 10,
	.icon_y = 10,
	.testwrite = 1,
	.del_mdb = 1,
	.del_empty_ts = 1,
	.io_buffer = 1732608,
};

struct pvr_enum_entry {
	char *name;
	int value;
};

static struct pvr_enum_entry ctl_unicode[] = {
	{"UTF",   CFG_UNICODE_UTF8},
	{"UTF-8", CFG_UNICODE_UTF8},
	{"UTF8",  CFG_UNICODE_UTF8},
	{"ASCII", CFG_UNICODE_ASCII},
	{"ASCII_TRANSCODE", CFG_UNICODE_ASCII_TRANSCODE},
	{NULL, 0}
};

static struct pvr_enum_entry ctl_default_source[] = {
	{"OFF",       0},
	{"TV",        KEY_SOURCE_TV},
	{"EXT1",      KEY_SOURCE_EXT1},
	{"EXT2",      KEY_SOURCE_EXT2},
	{"PC",        KEY_SOURCE_PC},
	{"AV",        KEY_SOURCE_AV},
	{"COMPONENT", KEY_SOURCE_COMPONENT},
	{"HDMI1",     KEY_SOURCE_HDMI1},
	{"HDMI2",     KEY_SOURCE_HDMI2},
	{"HDMI3",     KEY_SOURCE_HDMI3},
	{"DVI",       KEY_SOURCE_HDMI3},
	{"HDMI4",     KEY_SOURCE_HDMI4},
	{NULL, 0}
};

#define STR_ENTRY(str) fill_string, config.str, (void *)sizeof(config.str)
#define FLAG_ENTRY(flag) fill_flag, &config.flag, NULL
#define INT_ENTRY(integer) fill_int, &config.integer, NULL
#define ENUM_ENTRY(integer, ctl) fill_enum, &config.integer, ctl

static struct pvr_config_var {
	char *name;
	int (*fill_cb)(char *name, char *str, void *ptr, void *ctl);
	void *ptr;
	void *ctl;
} vars[] = {
	{"TARGET1",         STR_ENTRY(target1)},
	{"TARGET2",         STR_ENTRY(target2)},
	{"LOG",             FLAG_ENTRY(debug)},
	{"EXEDSP_LOG",      FLAG_ENTRY(exedsp_debug)},
	{"DEFAULT_SOURCE",  ENUM_ENTRY(default_source, ctl_default_source)},
	{"ASPECT_RATIO",    FLAG_ENTRY(aspect_ratio)},
	{"RECORD_KEY_LOCK", INT_ENTRY(record_key_lock)},
	{"PROTECT",         FLAG_ENTRY(protect)},
	{"UNICODE",         ENUM_ENTRY(unicode, ctl_unicode)},
	{"WINTER_TIME",     INT_ENTRY(winter_time)},
	{"SUMMER_TIME",     INT_ENTRY(summer_time)},
	{"SHOW_ICONS",      FLAG_ENTRY(show_icons)},
	{"ICON_X",          INT_ENTRY(icon_x)},
	{"ICON_Y",          INT_ENTRY(icon_y)},
	{"TESTWRITE",       FLAG_ENTRY(testwrite)},
	{"DEL_MDB",         FLAG_ENTRY(del_mdb)},
	{"DEL_EMPTY_TS",    FLAG_ENTRY(del_empty_ts)},
	{"IO_BUFFER",       INT_ENTRY(io_buffer)},
	{NULL, NULL, NULL, NULL}
};

static char *
skip_spaces(char *ptr)
{
	while (isspace(*ptr))
		ptr++;
	return (ptr);
}

static int
fill_int(char *name, char *str, void *ptr, void *ctl)
{
	int var;

	if ((var = atoi(str)) < 0)
		var = 0;

	*(unsigned *)ptr = var;
	LOG("%s = [%d]", name, var);
	
	return (0);
}

static int
fill_flag(char *name, char *str, void *ptr, void *ctl)
{
	int *flag = ptr;

	if (!strcasecmp(str, "yes") || !strcasecmp(str, "true") ||
		!strcasecmp(str, "enable") || !strcasecmp(str, "enabled") ||
		!strcasecmp(str, "on")) {

		*flag = 1;
	} else if (!strcasecmp(str, "no") || !strcasecmp(str, "false") ||
			   !strcasecmp(str, "disable") || !strcasecmp(str, "disabled") ||
			   !strcasecmp(str, "off")) {
		*flag = 0;
	} else if (*str == '\0')
		*flag = 0;
	else
		*flag = (atoi(str) > 0);

	LOG("%s = [%d]", name, *flag);

	return (0);
}

static int
fill_string(char *name, char *str, void *ptr, void *ctl)
{
	size_t size = (size_t)ctl;

	strlcpy((char *)ptr, str, size);
	LOG("%s = [%s]", name, (char *)ptr);

	return (0);
}

static int
fill_enum(char *name, char *str, void *ptr, void *ctl)
{
	struct pvr_enum_entry *entry = ctl;

	for (; entry->name != NULL; entry++) {
		if (strcasecmp(str, entry->name))
			continue;
		*(int *)ptr = entry->value;
		LOG("%s = [%d](%s)", name, entry->value, str);
		return (0);
	}
	LOG("Unknown %s value[%s], set default[%d]",
		name, str, *(int *)ptr);
	return (0);
}

void
load_config(const char *path)
{
	struct pvr_config_var *var;
	FILE *cfg;
	char buf[MAX_PATH];
	char *name, *val, *ptr;
	int i;

	strlcpy(config.path, path, sizeof(config.path));

	/*
     * Initialize the logging. It can be desabled later, in
     * pvr_config().
     */
	snprintf(buf, sizeof(buf), "%s%s", config.path, "pvr.log");
	pvr_log_init(buf);
	LOG("Loading configuration file[%s]", buf);

	snprintf(buf, sizeof(buf), "%s%s", path, CONFIG_FNAME);
	if ((cfg = fopen(buf, "r")) == NULL)
		return;

	while (fgets(buf, sizeof(buf), cfg) != NULL) {
		for (i = 0; buf[i] != '\0'; i++) {
			if (buf[i] != '\r' && buf[i] != '\n')
				continue;
			buf[i] = '\0';
			break;
		}
		name = skip_spaces(buf);

		/* skip comment lines or lines with unknown format */
		if (strchr(";#/=", *name) != NULL) {
			continue;
		}
		if ((val = strchr(name, '=')) == NULL) {
			continue;
		}

		for (ptr = val - 1; ptr >= name && isspace(*ptr); ptr--);
		ptr[1] = '\0';

		for (var = vars; var->name != NULL; var++)
			if (!strcasecmp(var->name, name))
				break;
		if (var->name == NULL) {
 			LOG("Unknown variable[%s]",	name);
			continue;
		}

		val = skip_spaces(val + 1);
		for (ptr = val + strlen(val) - 1; ptr >= val && isspace(*ptr); ptr--);
		ptr[1] = '\0';

		var->fill_cb(name, val, var->ptr, var->ctl);
	}

	fclose(cfg);
}
